
/*
 * ================================================================================
 *
 *     File        : foxy_http.c
 *     Author      : Michail Flouris <flouris@ics.forth.gr>
 *     Description : HTTP code file for the Foxy Web Cache.
 *
 * 
 *  Copyright (c) 1998-2001 Michail D. Flouris
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  URL: http://www.gnu.org/licenses/gpl.html
 * 
 *     NOTE : To view this code right use tabstop = 4.
 *     This code was written with Vi Improved (www.vim.org)
 * ================================================================================
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>

#include "foxy.h"

static char foxy_error_buf[FOXY_OS_PAGE_SIZE];	/* Global buffer for returning errors */
static char foxy_hdr[ FOXY_REQ_MAX_SIZE ];		/* HTTP Request Header reading buffer */

inline int
foxy_http_error( int conn, int error, char *msg )
{
	static char *msgskel = \
"HTTP/1.0 %d %s\r\n Content-type: text/html\r\n\r\n <html><head><title>%s</title></head>\r\n <body><p>\r\n <font size=\"+2\">Foxy Cache Error.</font><p>\r\n An error of type %d occurred: %s<p>\r\n <hr>\r\n <font size=\"-1\"><em>Generated by %s (C)1999-2001 Michail Flouris.</em></font>\r\n </body></html>\r\n";

	if ((foxy_conn[conn].clientfd <= 0) || (! msg))
		return 0;

	memset( foxy_error_buf, 0, FOXY_OS_PAGE_SIZE );
	sprintf(foxy_error_buf, msgskel, error, msg, msg, error, msg, FOXY_PROG_ID);
	write(foxy_conn[conn].clientfd, foxy_error_buf, strlen(foxy_error_buf));

	return 1;
}



inline int
read_and_parse_http_request( int conn )
{
	register int i, j, j1, k1, k2, k3, sl, eorl = 0;

	/* Clear the header buffer */
	memset(foxy_hdr, 0, FOXY_REQ_MAX_SIZE);

	/* Receive the request ... */
	if((j = read(foxy_conn[conn].clientfd, foxy_hdr, FOXY_REQ_MAX_SIZE)) == -1) {
		switch (errno) {
			case EWOULDBLOCK:
				foxy_notice( 0, "read_and_parse_http_request", 'i', "Socket Read : EWOULDBLOCK", 0);
				return -1;
			break;
			case EINTR:
				foxy_notice( 0, "read_and_parse_http_request", 'i', "Socket Read : EINTR", 0);
				return -1;
			break;
			default:
				foxy_notice( 0, "read_and_parse_http_request", 'i', "Socket Read : Default", 0);
				return 0;
			break;
		}
	}
	if (j <= 0)
		return 0;

		/* Locate the string "\r\n" in the foxy_hdr and parse
		 * the request line */
	j = j1 = k1 = k2 = k3 = 0;
	for (i = 0; i < FOXY_REQ_MAX_SIZE; i++) {
		if (foxy_hdr[i] == ' ') {
			if (!j)
				j = i;	/* Find the first 'space' */
			else { /* confusing ifs require { & } */
				if (!j1)
				j1 = i;	/* Find the second 'space' */
			}
		}
		/* Find the file pathname in the server ... */
		else if ((foxy_hdr[i] == '/') && (k1 < 3)) {
			switch (k1) {
			case 0:
				k1++;
			break;
			case 1:
				k2 = i;
				k1++;
			break;
			case 2:
				k1 = i;
			break;
			}
		}
		else if ((foxy_hdr[i] == ':') && (k3 < 2)) {
			if (k3 == 0)
				k3++;
			else
				k3 = i;
		}
		if ((foxy_hdr[i] == '\r') &&
			(foxy_hdr[i+1] == '\n')) {
			eorl = i+2;
			break;
		}
		if (foxy_hdr[i] == '\0')
			break;
	}

		/* Did not find \r\n in request line. This is against the HTTP RFC
		 * but we assume it exists in the end of the request. */
	if (i != eorl - 2) {
		eorl = i;
	}
	if (eorl >= FOXY_MAX_OBJNAME_LEN) {
		foxy_notice( 0, "read_and_parse_http_request", 'e', "Request string too long ...", 0);
		foxy_http_error( conn, 400, "Request string too long.");
		return 0;
	}

	if ( !j || !k1 || (j > 5)) {
		foxy_notice( 0, "read_and_parse_http_request", 'e', "Invalid request ...", 0);
		foxy_http_error( conn, 400, "Unrecognizable request");
		return 0;
	}
	/* We only do the "GET" method ... */
/*	if ( memcmp( &(foxy_hdr[0]), "GET", 3 ) ) {
		foxy_http_error( conn, 400, "Foxy only does GET method ... Sorry.");
		return 0;
	}*/
	/* Check if we have HTTP protocol > 1.0 ... */
	if ( memcmp( &(foxy_hdr[eorl-10]), "HTTP/1.1", 8 ) == 0 ) {
		foxy_http_error( conn, 400, "Foxy only does HTTP/1.0 and below ... Sorry.");
		foxy_notice( 0, "read_and_parse_http_request", 'e', "Protocol : %8s\n", 1,
							&(foxy_hdr[eorl-10]) );
		return 0;
	}
	/* We only do the http protocol for now ... */
	if ( memcmp( &(foxy_hdr[j+1]), "http://", 7 ) ) {
		foxy_http_error( conn, 400, "Foxy only does http protocol ... Sorry.");
		return 0;
	}
	sl = strlen(foxy_hdr);

	/* Check the memory space for the URLs (= objnames) */
	while ( foxy_connbuf[conn].objname_sz < (j1 - j) ) {
#ifdef __GNUC__
		foxy_notice( 0, __PRETTY_FUNCTION__, 'i', "Reallocating URL name space ...", 0);
#else
		foxy_notice( 0, "read_and_parse_http_request", 'i', "Reallocating URL name space ...", 0);
#endif
#ifdef FOXY_USE_MM
		foxy_connbuf[conn].idx_info.objname =
			(char *) foxy_mem_realloc( foxy_connbuf[conn].idx_info.objname,
									   foxy_connbuf[conn].objname_sz,
									   foxy_connbuf[conn].objname_sz + 128 );
#else
		foxy_connbuf[conn].idx_info.objname =
			(char *) foxy_realloc( foxy_connbuf[conn].idx_info.objname,
								   foxy_connbuf[conn].objname_sz,
								   foxy_connbuf[conn].objname_sz + 128 );
#endif
			 foxy_connbuf[conn].objname_sz += 128;
	}

	memset( foxy_connbuf[conn].req_line, 0, foxy_connbuf[conn].req_line_sz );
	memcpy( &(foxy_connbuf[conn].req_line[0]), &(foxy_hdr[0]), j );
	foxy_connbuf[conn].req_line[j] = ' ';
	memcpy( &(foxy_connbuf[conn].req_line[j+1]),
			&(foxy_hdr[k1]), eorl - k1 );
	memset( foxy_connbuf[conn].rest_header, 0, foxy_connbuf[conn].rest_header_sz );
	memcpy( foxy_connbuf[conn].rest_header,
			(const void *) &(foxy_hdr[eorl]), sl - eorl );
	if ( k1 - k2 > FOXY_DNS_MAX_HOSTNAME_LEN) {
		foxy_notice( 0, "read_and_parse_http_request", 'i', "Hostname too long ...", 0);
		return 0;
	}
	memset( foxy_connbuf[conn].srvhostname, 0, k1 - k2 );

	/* Copy the Object name (which is the absolute URL) in the index struct ... */
	/* CAUTION : The caller must have the foxy_connbuf[conn].idx_info.objname zero'ed */
	memset( foxy_connbuf[conn].idx_info.objname, 0, foxy_connbuf[conn].objname_sz);
		/* Cut off the trailing '/' */
	if ( foxy_hdr[j1-1] == '/' )
		memcpy( foxy_connbuf[conn].idx_info.objname,
				(const void *) &(foxy_hdr[j+1]), j1 - j - 2 );
	else
		memcpy( foxy_connbuf[conn].idx_info.objname,
				(const void *) &(foxy_hdr[j+1]), j1 - j - 1 );

	if ((k3 == 1) || (k1 - k3 <= 0)) {
		foxy_connbuf[conn].srvport = 80;
		memcpy( foxy_connbuf[conn].srvhostname,
				&(foxy_hdr[k2+1]), k1 - k2 - 1 );
	} else {
		char pbuf[6];
		memset(pbuf, 0, 6);
		memcpy( pbuf, &(foxy_hdr[k3+1]), k1 - k3 );
		foxy_connbuf[conn].srvport = atoi( pbuf );
		if (foxy_connbuf[conn].srvport <= 0) {
			foxy_notice( 0, "read_and_parse_http_request", 'e', "Port string : %s\n", 1, pbuf);
			foxy_http_error( conn, 503, "Invalid Port number");
			return 0;
		}
		memcpy( foxy_connbuf[conn].srvhostname,
				&(foxy_hdr[k2+1]), k3 - k2 - 1 );
	}

/*
	foxy_notice( 0, "read_and_parse_http_request", 'i', "foxy_connbuf[conn].req_line : %s\n", 1,
						foxy_connbuf[conn].req_line);
	foxy_notice( 0, "read_and_parse_http_request", 'i', "foxy_connbuf[conn].srvhostname : %s\n", 1,
						foxy_connbuf[conn].srvhostname);
	foxy_notice( 0, "read_and_parse_http_request", 'i', "foxy_connbuf[conn].srvport : %d\n", 1,
						foxy_connbuf[conn].srvport);
	foxy_notice( 0, "read_and_parse_http_request", 'i', "foxy_connbuf[conn].rest_header : %s\n", 1,
						foxy_connbuf[conn].rest_header);
*/

	return 1;
}


	/* Send the HTTP request to the Web Server ... */
inline int
send_http_request( int conn )
{
	int s, t;

	/* Use the global HTTP Header buffer to construct the header */
	memset( foxy_hdr, 0, FOXY_REQ_MAX_SIZE );
	s = strlen(foxy_connbuf[conn].req_line);
	t = s;
	memcpy( foxy_hdr, (char *) foxy_connbuf[conn].req_line, s );
	s = strlen(foxy_connbuf[conn].rest_header);
	memcpy( foxy_hdr+t, (char *) foxy_connbuf[conn].rest_header, s );

	/* Send the request ... */
	if ( write(foxy_conn[conn].serverfd, foxy_hdr, s + t) == -1) {
		foxy_notice( 0, "send_http_request", 'e', "Sending request : %s\n", 1,
							strerror(errno));
		foxy_http_error( conn, 404, "Unable to connect to host");
		foxy_unregister_connection( conn );
		return 0;
	}

	return 1;
}



	/* Get the URL from its Web Server asynchronously (may return
	 * before all data is read) ... */
inline int
async_recv_http_response( int conn )
{
	register int len;

	/* If it's the begining of the object's transfer, read the header and decide
	 * how big a buffer we need... then allocate such a buffer from the buffer pool. */
	if (foxy_connbuf[conn].idx_info.objsize == 0) {

		/* Read the object's HTTP header (max FOXY_REQ_MAX_SIZE) ... */
		memset(foxy_hdr, 0, FOXY_REQ_MAX_SIZE);
		len = read(foxy_conn[conn].serverfd, foxy_hdr, FOXY_REQ_MAX_SIZE); /* REQ_MAX_SIZE !!! not more */

	/* Do we need more space than foxy_config.max_object_size ?
	 * (THIS MEANS THE OBJECT IS NOT CACHEABLE because it's too big !!!
	 * So we overwrite the buffer from the start ... */
	} else if (foxy_connbuf[conn].idx_info.objsize < foxy_config.max_object_size) {

			if (foxy_connbuf[conn].idx_info.objsize + FOXY_SOCK_READ_SIZE >
													foxy_config.max_object_size) {

				/* Read whatever data is there (max max_object_size - objsize) ... */
				len = read(foxy_conn[conn].serverfd,
						 foxy_connbuf[conn].buffer.data + foxy_connbuf[conn].idx_info.objsize,
						 foxy_config.max_object_size - foxy_connbuf[conn].idx_info.objsize);
			} else {

				/* Read whatever data is there (max FOXY_SOCK_READ_SIZE) ... */
				len = read(foxy_conn[conn].serverfd,
						 foxy_connbuf[conn].buffer.data + foxy_connbuf[conn].idx_info.objsize,
						 FOXY_SOCK_READ_SIZE);
			}
		/* objsize > max_object_size ... */
	} else {

		/* Read whatever data is there (max max_object_size - objsize) ... */
		len = read(foxy_conn[conn].serverfd,
				 (char *) foxy_connbuf[conn].buffer.data, FOXY_SOCK_READ_SIZE);

	}

	/* First check the return value "len":
	 * A. Len < 0 (There was an error when reading from the Web server socket).
	 * Check the error flags for "EAGAIN" which means we should wait for data to arrive ... */
	if (len == -1) {

		if (errno == EAGAIN) /* No data has arrived. Wait ... */
			return 0;
			
		else {	/* A real error has occured */
			foxy_notice( 0, "async_recv_http_response", 'e', "Receiving URL data : %s\n", 1,
							strerror(errno));
			foxy_http_error( conn, 400, "Error contacting web server.");
			return -1;
		}

	/* B. Len == 0 (Object Retrieval Finished OK) */
	} else if (len == 0)
		return 1;

	/* C. Len > 0 (We read some data, but haven't finished the retrieval ...)
	 * When the buffer size is zero we have read the HTTP Reply from the server.
	 * We have to parse it and process it ... */
	if ((foxy_connbuf[conn].idx_info.objsize == 0) && (len > 0)) {
		register int i, j, k, l;
		int cacheflag = 0;

		if ( strncmp( foxy_hdr, "HTTP/1.0 200 OK\r\n", 17) != 0) {
			foxy_notice( 0, "async_recv_http_response", 'e', "Error Reply from Web Server. %s\n",
						 1, foxy_hdr);
			foxy_notice( 0, "async_recv_http_response", 'e', "URL: %s\n",
						 1, foxy_connbuf[conn].idx_info.objname);
			/* Send the Reply "as-is" to the client ... */
			write(foxy_conn[conn].clientfd, foxy_hdr, len);
			return -1;
		}

		/* Since the Reply was OK, parse the header ... Entering the Line Parser ... */
		i = j = 17;
		/* All objects are considered cacheable, until proven uncacheable :-) */
		while ( i < len ) {

			/* Mark one line or the end of the header ... */
			k = l = 0;
			for (j = i; (j < len) && (!l); j++) {
				switch ( foxy_hdr[j] ) {
				case ':':
					if ( !k )	k = j;	/* Mark the first ':' position */
				break;
				case '\r':
					/* The line has ended ... */
					if ( foxy_hdr[j+1] == '\n' ) {
						j++; l = 1;
					}
				break;
				}
			}

			/* End of HTTP Reply Header located ... Check if we have to add some headers ... */
			if (j - i == 2) {

				/* Check if we have a buffer */
				if ( foxy_connbuf[conn].buffer.number < 0 ) {
					foxy_notice( 0, "async_recv_http_response", 'e', "Null Buffer for object !\n", 0);
					foxy_http_error( conn, 503, "Internal Error");
					return -1;
				}

				/* Copy the header into the object buffer */
				memcpy( foxy_connbuf[conn].buffer.data, foxy_hdr, len );
				foxy_connbuf[conn].idx_info.objsize = len;

				/* DON'T TRY TO READ AGAIN HERE ... */

				return 0;	/* Success */

			} else if (j - i > FOXY_REQ_MAX_SIZE)	{	/* Something is wrong :-) */

				foxy_notice( 0, "async_recv_http_response", 'e', "TOO LONG HTTP HEADER !\n", 0);
				foxy_http_error( conn, 503, "Internal Error");
				return -1;	/* Failure ... */
			}

			/* Process the line now (parse the header fields) ... */
			switch ( k - i ) {		/* Compare according to the size of the "attribute name" */
			case 6:
				if ((j - i == 18) &&
					(strncmp( foxy_hdr+i, "Pragma: no-cache", 16 ) == 0 )) {
//					fprintf( stderr, "TOKEN: Pragma: no-cache\n");
					foxy_connbuf[conn].objattr = (char) FOXY_OBJATTR_UNCACHEABLE;
					cacheflag = 1;
				}
			break;
			case 13:
				/* Parse the "Cache-Control" attribute ... */
				if ( strncmp( foxy_hdr+i, "Cache-Control", 13 ) == 0 ) {
					switch ( j - i ) {
					case 23:
						/* We have a Cacheable Document ... */
						if ( strncmp( foxy_hdr+i, "Cache-Control: public", 21 ) == 0 ) {
//							fprintf( stderr, "TOKEN: Cache-Control: public\n");
							foxy_connbuf[conn].objattr = (char) FOXY_OBJATTR_CACHEABLE;
							cacheflag = 1;
						}
					break;
					case 33:
						/* We have an Un-Cacheable Document ... */
						if ( strncmp( foxy_hdr+i,
									  "Cache-Control: private,no-cache", 31 ) == 0 ) {
//							fprintf( stderr, "TOKEN: Cache-Control: private,no-cache\n");
							foxy_connbuf[conn].objattr = (char) FOXY_OBJATTR_UNCACHEABLE;
							cacheflag = 1;
						}
					break;
					}
				}
			break;
			case 14:
				/* Parse the "Cache-Control" attribute ... */
				if ( strncmp( foxy_hdr+i, "Content-Length", 14 ) == 0 ) {

					l = (int) strtol( foxy_hdr+i+16 , (char **)NULL, 10);
					if ((l == 0) || (l <= 0) || (l > 10485760)) {	/* Check the return value of strtol */
						foxy_notice( 0, "async_recv_http_response", 'e',
									 "Invalid Content-Length attribute: %d.\n", 1, l);
						foxy_http_error( conn, 503, "Internal Error");
						return -1;
					}

					/* Get a buffer of suitable size from the buffer pool to read the object in ...*/
					if (l+FOXY_REQ_MAX_SIZE < foxy_config.max_object_size) {
						if ( ! foxy_bufpool_request( l+FOXY_REQ_MAX_SIZE, &(foxy_connbuf[conn].buffer)) ) {
							foxy_notice( 0, "async_recv_http_response", 'e',
										 "Cannot allocate Buffer of %d bytes for Disk Read !", 1,
										 l+FOXY_REQ_MAX_SIZE);
							foxy_http_error( conn, 503, "Internal Error");
							return -1;
						}
					} else {
						foxy_notice( 0, "async_recv_http_response", 'w', "Object too large to buffer !", 0);
						if ( ! foxy_bufpool_request( foxy_config.max_object_size,
													 &(foxy_connbuf[conn].buffer)) ) {
							foxy_notice( 0, "async_recv_http_response", 'e',
										 "Cannot allocate Buffer of %d bytes for Disk Read !", 1,
										 foxy_config.max_object_size);
							foxy_http_error( conn, 503, "Internal Error");
							return -1;
						}
					}
				}
			break;
			}		/* end switch ... */

			i = j;	/* Move the "begin pointer" to the current "end pointer" ... */
		}	/* end while ( i < len )  ... */

	}	/* end if (foxy_connbuf[conn].idx_info.objsize == 0) ... */

	foxy_connbuf[conn].idx_info.objsize += len;

	return 0;
}


